/* ppcrack 0.4 - decrypts PowerPacker encrypted data files with brute force
 * by Stuart Caie <kyzer@4u.net>, this software is in the Public Domain
 *
 * Webpage: http://www.kyz.uklinux.net/packers.php#ppcrack
 *
 * Usage: ppcrack <files>
 *        ppcrack -pass <key> <files>
 *        ppcrack -key <key> <files>
 *        ppcrack -from <key> -to <key> <files>
 *
 * Examples: ppcrack -key 034b5769 file.pp
 *           ppcrack -from 00000000 -to 1fffffff file.pp
 *           ppcrack -pass MyPassword *.pp
 *           ppcrack *.pp
 *
 * Anything which decrypts then decrunches to valid data is saved to disk
 * as <original filename>.<decryption key>
 *
 * As a bonus, if any file is a PowerPacker data file, but not encrypted,
 * it will be decrunched anyway, and saved as <original filename>.decrunched
 */

/* version history
 * 0.1  09-Dec-2001 - original release
 * 0.2  25-Nov-2002 - fixed off-by-one error detection error in PP_BYTE_OUT()
 *                  - allowed decrunching to end after a literal run.
 *                  - now ppValidate doesn't need the buffer decrypted.
 *                  - new command-line interface
 * 0.3  07-Sep-2003 - vastly improved skipping of known incorrect keys
 *                  - display update now includes keys/second
 *                  - improved documentation
 * 0.4  18-Sep-2003 - smaller command line parser
 *                  - support for PPLS files
 *                  - now writes key update / fatal errors to stderr
 *                  - split the basic PowerPacker functions into pplib.c
 *                  - support large files that aren't fully encrypted (!)
 *                  - improved documenatation
 * 0.4b 18-Sep-2003 - added "-big" option; the big files problem is only
 *                    in some unofficial versions of powerpacker.library.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <pplib.h>

/* globals */
#define OUTPUT_NAME_LEN (4096)
char output_name[OUTPUT_NAME_LEN];
unsigned int startkey, endkey, big_mode;


size_t filelen;
void *loadfile(char *name) {
  void *mem = NULL; FILE *fd;
  if ((fd = fopen(name, "rb"))) {
    if ((fseek(fd, 0, SEEK_END) == 0) && (filelen = ftell(fd))
    &&  (fseek(fd, 0, SEEK_SET) == 0) && (mem = malloc(filelen))) {
      if (fread(mem, 1, filelen, fd) < filelen) { free(mem); mem = NULL; }
    }
    fclose(fd);
  }
  return mem;
}
int savefile(char *name, void *mem, size_t length) {
  FILE *fd = fopen(name, "wb");
  int ok = fd && (fwrite(mem, 1, length, fd) == length);
  if (fd) fclose(fd);
  return ok;
}

int save_decrunched(char *name, unsigned char *offsets, unsigned char *dest,
		    unsigned int srclen, unsigned int destlen,
		    unsigned int key)
{
  unsigned char *src = &offsets[4];
  unsigned int srclen_bug;
  int result = 0;

  if (big_mode) {
    srclen_bug = srclen + 3;
    srclen_bug = (srclen_bug >> 2) - 1;
    srclen_bug &= 0xFFFF;
    srclen_bug = (srclen_bug + 1) << 2;
  }
  else {
    srclen_bug = srclen;
  }

  /* decrypt with this key */
  ppDecrypt(src, srclen_bug, key);
  if (ppDecrunchBuffer(offsets, dest, srclen, destlen)) {
    printf("%s: key %08X success!\n", name, key);
    result = 1;
  }
  else {
    printf("%s: decrunch failed, file is corrupt.\n", name);
  }
  fflush(stdout);

  /* generate filename and save */
  snprintf(output_name, OUTPUT_NAME_LEN, "%s.%08X", name, key);
  savefile(output_name, dest, destlen);

  /* re-encrypt for next run */
  ppDecrypt(src, srclen_bug, key);
  return result;
}

/* ------------------------------------------------------------------------- */

/* READ_BITS_SMALL(nbits, var) performs direct decryption of the data
 * stream while reading. The key is stored in key_array[], a 4 byte
 * array. The pointer kp is predecremented to the byte of the key to be
 * XORed with the source data.
 */
#define READ_BITS_SMALL(nb, v) do {                        \
  for (bc = (nb); bl < bc; bl += 8) {                      \
    if (buf < src) goto done; /* out of source bits */     \
    bb |= ((*--buf ^ *--kp) << bl); if (kp==ka) kp=&ka[4]; \
  }                                                        \
  (v)=0; bl-=bc; while(bc--){(v)<<=1; (v)|=bb&1; bb>>=1;}  \
} while (0)

unsigned int ppcrack_small(char *name, unsigned char *src,
  unsigned char *dest, unsigned int srclen, unsigned int destlen)
{
  unsigned int key, keys_todo, drop_mask, drop_key, time_mask, success;
  unsigned char ka[4], *buf, *kp, *offsets, skip_bits, bl, bc;
  unsigned int written, bb, x, y, ob, offset;

  time_t time_start, time_then, time_now;

  offsets = src; src += 4;
  skip_bits = src[srclen+3];

  /* Halving the keyspace from 2^32 to 2^31:
   *
   * PowerPacker alternates between two operations - literal and match.
   * The FIRST operation must be a literal, as there is no data to match
   * yet, so the first bit in the compressed stream must be a 0. The '8
   * bits other info' is the number of bits unused in the first
   * longword. These are ignored.
   *
   * We know the encrypted value of the first bit in the stream, and we
   * know it must actually be 0 when XORed. So that bit of they key must
   * equal the encrypted bit. We can reject all keys are the opposite of
   * this bit.
   */
  drop_mask = 1 << skip_bits;
  drop_key = ~((src[srclen-4] << 24) | (src[srclen-3] << 16) |
	       (src[srclen-2] << 8)  |  src[srclen-1]        ) & drop_mask;

  /* if start key or end key are in a drop range, move them out */
  if ((startkey & drop_mask) == drop_key) {
    startkey |= drop_mask-1;
    startkey++;
  }
  if ((endkey & drop_mask) == drop_key) {
    endkey &= ~(drop_mask-1);
    endkey--;
  }

  if (startkey > endkey) {
    printf("%s: no key could decrypt the file.\n", name);
    return 0;
  }
  key = startkey;
  keys_todo = endkey - startkey;

  time_start = time_then = time(NULL);
  time_mask = 0;
  success = 0;

  printf("%s: key bit %d=%d, searching %08X-%08X (%u keys)\n", name,
	 skip_bits, drop_key ? 0 : 1, startkey, endkey,
	 (keys_todo == 0xFFFFFFFF) ? keys_todo : keys_todo+1);

  /* main key search loop */
  do {
    if ((key & drop_mask) == drop_key) {
      key       += drop_mask;
      keys_todo -= drop_mask;
      continue;
    }

    /* display current key / key rate */
    if ((keys_todo & time_mask) == 0) {
      /* regulate the time_mask to be roughly once a second */
      if ((time_now = time(NULL)) == time_then) {
	time_mask = (time_mask << 1) | 1;
      }
      time_then = time_now;
      if (time_now != time_start) {
	fprintf(stderr, "key %08X (%u key/s)    \r", key,
		(key - startkey) / (time_now - time_start));
	fflush(stderr);
      }
    }

    /* attempt current key */
    ka[0] = key >> 24; /* uses implicit '& 0xFF' for speed */
    ka[1] = key >> 16;
    ka[2] = key >> 8;
    ka[3] = key;
    written = 0; buf = &src[srclen]; kp = &ka[4]; bl = 0; bb = 0;
    READ_BITS_SMALL(skip_bits, x);
    while (written < destlen) {
      READ_BITS_SMALL(1, x);
      if (x == 0) {
	y = 1; do { READ_BITS_SMALL(2, x); y += x; } while (x == 3);
	written += y; if (written > destlen) break;
	while (y--) READ_BITS_SMALL(8, x); if (written == destlen) break;
      }
      READ_BITS_SMALL(2, x); ob = offsets[x]; y = x + 2;
      if (x == 3) {
	READ_BITS_SMALL(1, x); if (x==0) ob=7; READ_BITS_SMALL(ob, offset);
	do { READ_BITS_SMALL(3, x); y += x; } while (x == 7);
      } else { READ_BITS_SMALL(ob, offset); }
      if (offset >= written) break;
      written += y; if (written > destlen) break;
    }
  done:
    if (written == destlen) {
      success += save_decrunched(name, offsets, dest, srclen, destlen, key);
    }

    key++;
  } while (keys_todo--);
  return success;
}

/* ------------------------------------------------------------------------- */

#define READ_BITS_NONE(nb, v) do {                         \
  for (bc = (nb); bl < bc; bl += 8) {                      \
    if (buf < src) goto done1; /* out of source bits */    \
    if (!ec--) {save_ok = 1; goto done1;} /* exit */       \
    bb |= *--buf << bl; kp--; if (kp == ka) kp = &ka[4];   \
  }                                                        \
  (v)=0; bl-=bc; while(bc--){(v)<<=1; (v)|=bb&1; bb>>=1;}  \
} while (0)

#define READ_BITS_LARGE(nb, v) do {                        \
  for (bc = (nb); bl < bc; bl += 8) {                      \
    if (buf < src) goto done2; /* out of source bits */    \
    if (!ec--) {ka[0] = key >> 24; ka[1] = key >> 16;      \
                ka[2] = key >> 8;  ka[3] = key; }          \
    bb |= ((*--buf ^ *--kp) << bl); if (kp==ka) kp=&ka[4]; \
  }                                                        \
  (v)=0; bl-=bc; while(bc--){(v)<<=1; (v)|=bb&1; bb>>=1;}  \
} while (0)

unsigned int ppcrack_large(char *name, unsigned char *src,
  unsigned char *dest, unsigned int srclen, unsigned int destlen)
{
  unsigned int written, encr, bb, x, y, ob, offset, ec;
  unsigned int key, keys_todo, time_mask, success;
  unsigned char ka[4], *buf, *kp, *offsets, skip_bits, bl, bc;
  time_t time_start, time_then, time_now;


  /* variables to save decruncher state */
  unsigned char *save_buf = &src[srclen], *save_kp, save_bl;
  unsigned int save_bb=0, save_ec, save_written, save_ok;

  offsets = src; src += 4;
  skip_bits = src[srclen+3];

  /* how many bytes are encrypted? */
  encr = srclen + 3;
  encr = (encr >> 2) - 1;
  encr &= 0xFFFF;
  encr = (encr + 1) << 2;

  key = startkey;
  keys_todo = endkey - startkey;

  time_start = time_then = time(NULL);
  time_mask = 0;
  success = 0;


  /* initial decompression of unencrypted section */
  written = 0; save_ok = 0; buf = &src[srclen];
  kp = &ka[4]; bl = 0; bb = 0; ec = srclen - encr;
  READ_BITS_NONE(skip_bits, x);
  while (written < destlen) {
    /* save top-of-loop state */
    save_buf = buf; save_bl = bl; save_bb = bb;
    save_kp = kp, save_ec = ec; save_written = written;
    READ_BITS_NONE(1, x);
    if (x == 0) {
      y = 1; do { READ_BITS_NONE(2, x); y += x; } while (x == 3);
      written += y; if (written > destlen) break;
      while (y--) READ_BITS_NONE(8, x); if (written == destlen) break;
    }
    READ_BITS_NONE(2, x); ob = offsets[x]; y = x + 2;
    if (x == 3) {
      READ_BITS_NONE(1, x); if (x==0) ob=7; READ_BITS_NONE(ob, offset);
      do { READ_BITS_NONE(3, x); y += x; } while (x == 7);
    } else { READ_BITS_NONE(ob, offset);}
    if (offset >= written) break;
    written += y; if (written > destlen) break;
  }
 done1:

  if (!save_ok) {
    printf("%s: file is corrupt\n", name);
    return 0;
  }

  printf("%s: large file, %u bytes encrypted, searching %08X-%08X (%u keys)\n",
	 name, encr, startkey, endkey,
	 (keys_todo == 0xFFFFFFFF) ? keys_todo : keys_todo+1);

  /* main key search loop */
  do {
    /* display current key / key rate */
    if ((keys_todo & time_mask) == 0) {
      /* regulate the time_mask to be roughly once a second */
      if ((time_now = time(NULL)) == time_then) {
	time_mask = (time_mask << 1) | 1;
      }
      time_then = time_now;
      if (time_now != time_start) {
	fprintf(stderr, "key %08X (%u key/s)    \r", key,
		(key - startkey) / (time_now - time_start));
	fflush(stderr);
      }
    }

    /* attempt current key -- start unencrypted, READ_BITS_LARGE will
     * write the key in when ec==0
     */
    ka[0] = ka[1] = ka[2] = ka[3] = 0;
    written = save_written; buf = save_buf;
    kp = save_kp; bl = save_bl; bb = save_bb; ec = save_ec;

    while (written < destlen) {
      READ_BITS_LARGE(1, x);
      if (x == 0) {
	y = 1; do { READ_BITS_LARGE(2, x); y += x; } while (x == 3);
	written += y; if (written > destlen) break;
	while (y--) READ_BITS_LARGE(8, x); if (written == destlen) break;
      }
      READ_BITS_LARGE(2, x); ob = offsets[x]; y = x + 2;
      if (x == 3) {
	READ_BITS_LARGE(1, x); if (x==0) ob=7; READ_BITS_LARGE(ob, offset);
	do { READ_BITS_LARGE(3, x); y += x; } while (x == 7);
      } else { READ_BITS_LARGE(ob, offset);}
      if (offset >= written) break;
      written += y; if (written > destlen) break;
    }
  done2:
    if (written == destlen) {
      success += save_decrunched(name, offsets, dest, srclen, destlen, key);
    }
    key++;
  } while (keys_todo--);
  return success;
}

/* ------------------------------------------------------------------------- */

void ppcrack(char *name, unsigned char *src, unsigned int srclen) {
  unsigned char *dest;
  unsigned int x, eff;

  x = (src[0] << 24) | (src[1] << 16) | (src[2] << 8) | src[3];
  switch (x) {
  case 0x50503230: /* PP20 */
    eff = 4;
    break;
  case 0x50504C53: /* PPLS */
    eff = 8;
    break;
  case 0x50583230: /* PX20 */
    eff = 6;
    printf("%s: password checksum = %04X\n", name, (src[4] << 8) | src[5]);
    break;
  default:
    printf("%s: file does not have the PP signature\n", name);

  }
  if ((srclen - eff) < 12) {
    printf("%s: file is too short\n", name);
    return;
  }
  if ((srclen - eff) & 0x03) {
    printf("%s: file has bad length\n", name);
    return;
  }

  x = (src[srclen-4] << 16) | (src[srclen - 3] << 8) | src[srclen - 2];
  printf("%s: decrunched length = %u bytes\n", name, x);
  if (!(dest = malloc(x))) {
    printf("%s: out of memory!\n", name);
    return;
  }

  if (eff == 6) {
    /* PX20 -- have to crack it */
    unsigned int success;

    if (startkey == endkey) {
      /* single key */
      success = save_decrunched(name, &src[eff], dest, srclen-(eff+8), x,
				startkey);
      printf("%s: the key %08X %s the file.\n",
	     name, startkey, (success) ? "correctly decrypted" :
			                 "did not decrypt");
    }
    else {
      time_t t = time(NULL);
      
      if (big_mode && ((srclen - eff - 8) > 262144))
	/* larger files may be partially unencrypted */
	success = ppcrack_large(name, &src[eff], dest, srclen-(eff+8), x);
      else
	success = ppcrack_small(name, &src[eff], dest, srclen-(eff+8), x);

      /* inform the user of the success */
      t = time(NULL) - t; if (!t) t=1;
      printf("%s: %d successful key(s) found between %08X and %08X.\n",
	     name, success, startkey, endkey);
      printf("%s: completed %u keys in %u seconds; %u keys/s\n",
	     name, endkey - startkey, t, (endkey - startkey) / t);
    }
  }
  else {
    /* not PX20, just decrunch it */
    printf("%s: not encrypted, decrunching anyway\n", name);
    if (ppDecrunchBuffer(&src[eff], dest, srclen - (eff+8), x)) {
      printf("%s: decrunch successful!\n", name);
      snprintf(output_name, OUTPUT_NAME_LEN, "%s.decrunched", name);
      savefile(output_name, dest, x);
    }
    else {
      printf("%s: decrunch failed, file is corrupt.\n", name);
    }
  }
  fflush(stdout);
  free(dest);
}

/* ------------------------------------------------------------------------- */

#define MYARG_KEY  (0)
#define MYARG_FROM (1)
#define MYARG_TO   (2)
#define MYARG_PASS (3)
#define MYARG_BIG  (4)
#define MYARG_UNK  (5)
const char *argn[] = {"key", "from", "to", "pass", "big"};

int main(int argc, char *argv[]) {
  int i, keymode = 0, passmode = 0, rangemode = 0;

  big_mode = 0;
  startkey = 0x00000000;
  endkey   = 0xFFFFFFFF;

  if (argc < 2) {
    fprintf(stderr, "Usage: %s [-big] [-key <key>] <file(s)>\n"
	            "   or  %s [-big] [-pass <password>] <file(s)>\n"
	            "   or  %s [-big] [-from <key>] [-to <key>] <file(s)>\n"
	            "Example: %s -key 30a9f3e3 secret.pp\n",
	    argv[0], argv[0], argv[0], argv[0]);
    return 1;
  }

  /* parse any '-key', '-pass', '-from' and '-to' options */
  for (i = 1; i < argc; i++) {
    unsigned int myarg = MYARG_UNK, j;
    if (argv[i][0] != '-') continue;
    for (j = 0; j < 5; j++) if (strcmp(&argv[i][1], argn[j]) == 0) myarg = j;
    switch (myarg) {
    case MYARG_KEY:                  keymode   = 1; break;
    case MYARG_FROM: case MYARG_TO:  rangemode = 1; break;
    case MYARG_PASS:                 passmode  = 1; break;
    case MYARG_BIG:                  big_mode  = 1; break;
    default:
      fprintf(stderr, "Unknown option '-%s'\n", &argv[i][1]);
      return 1;
    }
    if ((keymode + passmode + rangemode) > 1) {
      fprintf(stderr, "Only use one of -key, -pass or -from/-to\n");
      return 1;
    }
    if (myarg != MYARG_BIG) {
      if (!argv[++i]) {
	fprintf(stderr, "Missing argument for -%s option.\n", argn[myarg]);
	return 1;
      }

      if (myarg == MYARG_PASS) {
	startkey = endkey = ppCalcPasskey((unsigned char *) argv[i]);
	printf("Password '%s' gives checksum %04X, key %08X\n", argv[i],
	       ppCalcChecksum((unsigned char *) argv[i]), startkey);
      }
      else {
	if (sscanf(argv[i], "%8x", &j) != 1) {
	  fprintf(stderr, "Bad hexadecimal number '%s' for -%s option.\n",
		  argv[i], argn[myarg]);
	  return 1;
	}
	switch (myarg) {
	case MYARG_FROM: startkey          = j; break;
	case MYARG_TO:   endkey            = j; break;
	case MYARG_KEY:  startkey = endkey = j; break;
	}
      }
    }
  }

  if (startkey > endkey) {
    fprintf(stderr, "The start key (%08X) is more than the end key (%08X).\n",
           startkey, endkey);
    return 1;
  }

  if (startkey == endkey) {
    printf("Using only the key %08X to decrypt files.\n", startkey);
  }
  else {
    printf("Decrypting with keys from %08X to %08X.\n", startkey, endkey);
  }

  for (i = 1; i < argc; i++) {
    unsigned char *data, j;

    /* skip options */
    if (argv[i][0] == '-') {
      /* all valid options except '-big' have a parameter */
      for (j = 0; j < 4; j++) if (strcmp(&argv[i][1],argn[j])==0) {i++; break;}
      continue;
    }

    if ((data = loadfile(argv[i])) != NULL) {
      ppcrack(argv[i], data, (unsigned int) filelen);
      free(data);
    }
    else {
      perror(argv[i]);
    }
  }
  return 0;
}
